{
  The simple demo of sound processing
  Author:  Eugen Tarasov

  There is implementation of audio stream support.
  This unit is based on the DelphiX library.

  It contains:

  TdspAudioStream - play sound from any stream.

  TdspWaveFileStream - stream for working with the wav-files.
    The heard of this class is notification event to process
    a piece of sound data.

  TdspAudioFileStream - the derived class from TdspAudioStream.
    This class implements the sound playing from wav-file.
    This is the heard of our sound engine :)

  Last modified - January, 2004

  You can use this module on you own risk.
  Please send all suggestions to author

  e-mail: dsplab@teworks.com
  http://www.teworks.com
}

unit AudioStream;

interface

uses
  Forms, SysUtils, Windows, Wave, Classes, DXSounds, MMSystem, Math,
  DXConsts, DirectX, DXClass;

type
  TOnDataProcessing = procedure(var Buffer; Count: Longint) of object;

  TdspAudioStream = class
  private
    // DirectSound
    FDSound: TDXSound;
    // Direct Sound buffer
    FDSBuffer: TDirectSoundBuffer;
    FDSBufferSize: DWORD;
    FBufferLength: Integer;
    FBufferPos: DWORD;
    FPlayBufferPos: DWORD;
    FPlayedSize: Integer;
    FPlaying: Boolean;
    FPosition: Integer;
    FWaveStream: TCustomWaveStream;
    FWritePosition: Integer;
    FNotifyEvent: THandle;
    FNotifyThread: TThread;
    function GetFormat: PWaveFormatEX;
    function GetFrequency: Integer;
    function GetPan: Integer;
    function GetPlayedSize: Integer;
    function GetSize: Integer;
    function GetVolume: Integer;
    function GetWriteSize: Integer;
    procedure SetBufferLength(Value: Integer);
    procedure SetFrequency(Value: Integer);
    procedure SetPan(Value: Integer);
    procedure SetPlayedSize(Value: Integer);
    procedure SetPosition(Value: Integer);
    procedure SetVolume(Value: Integer);
    procedure SetWaveStream(Value: TCustomWaveStream);
    procedure Update2(InThread: Boolean);
    procedure UpdatePlayedSize;
    function WriteWave(WriteSize: Integer): Integer;
    procedure RecreateBuf;

  private
    FdspBufferSize: Longint;
    FdspBufferPos: Longint;
    FdspBuffer: Pointer;
    FOnDataProcessing: TOnDataProcessing;

    procedure FSetdspBufferSize(NewSize: Integer);

  protected
    function  Read(var Data; DataSize: Longint): Longint;

  public
    constructor Create(ADXSound: TDXSound);
    destructor Destroy; override;

    procedure Play;
    procedure Stop;

    property BufferLength: Integer read FBufferLength write SetBufferLength;
    property Format: PWaveFormatEx read GetFormat;

    property Volume: Integer read GetVolume write SetVolume;
    property Frequency: Integer read GetFrequency write SetFrequency;
    property Pan: Integer read GetPan write SetPan;
    property Position: Integer read FPosition write SetPosition;

    property Size: Integer read GetSize;
    property PlayedSize: Integer read GetPlayedSize write SetPlayedSize;
    property Playing: Boolean read FPlaying;

    property WaveStream: TCustomWaveStream read FWaveStream write SetWaveStream;
    property dspBufferSize: Longint read FdspBufferSize write FSetdspBufferSize;

    property OnDataProcessing: TOnDataProcessing read FOnDataProcessing write FOnDataProcessing;
  end;

  {  TdspAudioFileStream  }

  TdspAudioFileStream = class(TdspAudioStream)
  private
    FFileName: string;
    FWaveFileStream: TWaveFileStream;

    procedure SetFileName(const Value: string);
  public
    destructor Destroy; override;
    property FileName: string read FFileName write SetFileName;
  end;


implementation

{  TdspAudioStreamNotify  }

type
  TdspAudioStreamNotify = class(TThread)
  private
    FAudio: TdspAudioStream;
    FSleepTime: Integer;
  public
    constructor Create(Audio: TdspAudioStream);
    destructor Destroy; override;
    procedure Execute; override;
    procedure Update;
  end;


constructor TdspAudioStreamNotify.Create(Audio: TdspAudioStream);
begin
  inherited Create(True);
  FreeOnTerminate:= True;
  FAudio:= Audio;
  FAudio.FNotifyEvent:= CreateEvent(nil, False, False, nil);
  FSleepTime:= 50;
  Resume;
end;


destructor TdspAudioStreamNotify.Destroy;
begin
  FAudio.Stop;
  CloseHandle(FAudio.FNotifyEvent);
  inherited Destroy;
end;


procedure TdspAudioStreamNotify.Execute;
begin
  while WaitForSingleObject(FAudio.FNotifyEvent, FSleepTime) = WAIT_TIMEOUT do Synchronize(Update);
end;

procedure TdspAudioStreamNotify.Update;
begin
  try
    FAudio.Update2(True);
  except
    on E: Exception do
    begin
      Application.HandleException(E);
      SetEvent(FAudio.FNotifyEvent); // Terminate Thread
    end;
  end;
end;


{  TdspAudioStream  }

constructor TdspAudioStream.Create(ADXSound: TDXSound);
begin
  inherited Create;
  FDSound:= ADXSound;
  FDSBuffer:= TDirectSoundBuffer.Create(FDSound.DSound);
  BufferLength:= 1000;
  dspBufferSize:= 512 * 4;
end;


destructor TdspAudioStream.Destroy;
begin
  Stop;
  FDSBuffer.Free;
  dspBufferSize:= 0;
  inherited Destroy;
end;


procedure TdspAudioStream.FSetdspBufferSize(NewSize: Integer);
begin
  if NewSize <> FdspBufferSize then
  begin
    ReallocMem(FdspBuffer, NewSize);
    FdspBufferSize:= NewSize;
    FdspBufferPos:= FdspBufferSize;
  end;
end;


function TdspAudioStream.GetFormat: PWaveFormatEX;
begin
  if FWaveStream = nil then Result:= nil
  else Result:= WaveStream.Format;
end;


function TdspAudioStream.GetFrequency: Integer;
begin
  Result:= FDSBuffer.Frequency;
end;


procedure TdspAudioStream.SetFrequency(Value: Integer);
begin
  FDSBuffer.Frequency:= Value;
end;


function TdspAudioStream.GetPan: Integer;
begin
  Result:= FDSBuffer.Pan;
end;


procedure TdspAudioStream.SetPan(Value: Integer);
begin
  FDSBuffer.Pan:= Value;
end;


function TdspAudioStream.GetPlayedSize: Integer;
begin
  if Playing then UpdatePlayedSize;
  Result:= FPlayedSize;
end;


procedure TdspAudioStream.SetPlayedSize(Value: Integer);
begin
  if Playing then UpdatePlayedSize;
  FPlayedSize:= Value;
end;


function TdspAudioStream.GetVolume: Integer;
begin
  Result:= FDSBuffer.Volume;
end;


procedure TdspAudioStream.SetVolume(Value: Integer);
begin
  FDSBuffer.Volume:= Value;
end;


procedure TdspAudioStream.SetPosition(Value: Integer);
begin
  if not Assigned(FWaveStream) then Exit;

  Value:= Max(Min(Value, Size-1), 0);
  Value:= Value div Format^.nBlockAlign * Format^.nBlockAlign;

  FPosition:= Value;

  if Playing then
  begin
    try
      FDSBuffer.Stop;

      FBufferPos:= 0;
      FPlayBufferPos:= 0;
      FWritePosition:= Value;

      FdspBufferPos:= FdspBufferSize;
      FWaveStream.Position:= FWritePosition;

      WriteWave(FDSBufferSize);

      FDSBuffer.Position:= 0;
      FDSBuffer.Play(True);
    except
      Stop;
      raise;
    end;
  end;
end;


function TdspAudioStream.GetSize: Integer;
begin
 if Assigned(FWaveStream) then Result:= WaveStream.Size else Result:= 0;
end;


procedure TdspAudioStream.UpdatePlayedSize;
var PlayPosition, PlayedSize: DWORD;
begin
  PlayPosition:= FDSBuffer.Position;

  if FPlayBufferPos <= PlayPosition then PlayedSize:= PlayPosition - FPlayBufferPos
  else PlayedSize:= PlayPosition + (FDSBufferSize - FPlayBufferPos);

  Inc(FPlayedSize, PlayedSize);
  FPlayBufferPos:= PlayPosition;
end;


procedure TdspAudioStream.Play;
begin
  if not FPlaying then
  begin
    if FWaveStream = nil then raise Exception.Create(SWaveStreamNotSet);
    if Size = 0 then Exit;

    FPlaying:= True;
    try
      SetPosition(FPosition);
      FNotifyThread:= TdspAudioStreamNotify.Create(Self);
    except
      Stop;
      raise;
    end;
  end;
end;


procedure TdspAudioStream.Stop;
begin
  if FPlaying then
  begin
    FPlaying:= False;
    SetEvent(FNotifyEvent); // Terminate Notification Thread
    FDSBuffer.Stop;
  end;
end;


procedure TdspAudioStream.RecreateBuf;
var
  APlaying: Boolean;
  APosition: Integer;
  AFrequency: Integer;
  APan: Integer;
  AVolume: Integer;
begin
  APlaying:= Playing;

  APosition:= Position;
  AFrequency:= Frequency;
  APan:= Pan;
  AVolume:= Volume;

  SetWaveStream(WaveStream);

  Position:= APosition;
  Frequency:= AFrequency;
  Pan:= APan;
  Volume:= AVolume;

  if APlaying then Play;
end;


procedure TdspAudioStream.SetBufferLength(Value: Integer);
begin
  Value:= Max(10, Value);
  if FBufferLength <> Value then
  begin
    FBufferLength:= Value;
    if FWaveStream <> nil then RecreateBuf;
  end;
end;


procedure TdspAudioStream.SetWaveStream(Value: TCustomWaveStream);
var BufferDesc: TDSBufferDesc;
begin
  Stop;

  FWaveStream:= nil;
  FBufferPos:= 0;
  FPosition:= 0;
  FWritePosition:= 0;

  if (Value <> nil) and (FBufferLength > 0) then
  begin
    FDSBufferSize:= (FBufferLength * Integer(Value.Format^.nAvgBytesPerSec) div 1000);
    FDSBufferSize:= (FDSBufferSize div 4) * 4;

    FillChar(BufferDesc, SizeOf(BufferDesc), 0);
    with BufferDesc do
    begin
      dwSize:= SizeOf(TDSBufferDesc);
      dwFlags:= DSBCAPS_CTRLDEFAULT or DSBCAPS_GETCURRENTPOSITION2;
      if soStickyFocus in FDSound.NowOptions then dwFlags:= dwFlags or DSBCAPS_STICKYFOCUS else
      if soGlobalFocus in FDSound.NowOptions then dwFlags:= dwFlags or DSBCAPS_GLOBALFOCUS;
      dwBufferBytes:= FDSBufferSize;
      lpwfxFormat:= Value.Format;
      lpwfxFormat.wBitsPerSample:= 16;
    end;

    if not FDSBuffer.CreateBuffer(BufferDesc) then
      raise EDirectSoundBufferError.CreateFmt(SCannotMade, [SDirectSoundBuffer]);
  end else
  begin
    FDSBuffer.IDSBuffer := nil;
    FDSBufferSize := 0;
  end;

  FWaveStream := Value;
end;

function TdspAudioStream.GetWriteSize: Integer;
var
  PlayPosition: DWORD;
begin
  PlayPosition := FDSBuffer.Position;

  if FBufferPos <= PlayPosition then Result := PlayPosition - FBufferPos
  else Result := PlayPosition + (FDSBufferSize - FBufferPos);

  Result:= Result - (Result mod 4);
end;

procedure TdspAudioStream.Update2(InThread: Boolean);
var
  WriteSize: Integer;
begin
  if not FPlaying then Exit;

  try
    UpdatePlayedSize;

    if Size<0 then
    begin
      WriteSize := GetWriteSize;
      if WriteSize>0 then
      begin
        WriteSize := WriteWave(WriteSize);
        FPosition := FPosition + WriteSize;
      end;
    end else
    begin
      if FPosition<Size then
      begin
        WriteSize := GetWriteSize;
        if WriteSize>0 then
        begin
          WriteWave(WriteSize);
          FPosition := FPosition + WriteSize;
          if FPosition>Size then FPosition := Size;
        end;
      end else
      begin
        if InThread then
          SetEvent(FNotifyEvent)
        else
          Stop;
      end;
    end;
  except
    if InThread then
      SetEvent(FNotifyEvent)
    else
      Stop;
    raise;
  end;
end;


function  TdspAudioStream.Read(var Data; DataSize: Longint): Longint;
var
  ReadBytes: Longint;
  BytesToCopy: Longint;
  SrcBuf, DestBuf: PChar;
begin
  Result:= 0;
  DestBuf:= @Data;
  while DataSize > 0 do
  begin
    //1. LoadBuffer
    if FdspBufferPos >= FdspBufferSize then
    begin
      FdspBufferPos:= 0;
      ReadBytes:= FWaveStream.Read(FdspBuffer^, FdspBufferSize);

      if ReadBytes < FdspBufferSize then
        FillChar(
          Pointer(Integer(FdspBuffer) + ReadBytes)^,
          FdspBufferSize - ReadBytes, 0);

      //2. ProcessBuffer
      if Assigned(FOnDataProcessing) then FOnDataProcessing(FdspBuffer^, FdspBufferSize);
    end;

    //3. CopyToExternalBuffer
    SrcBuf:= FdspBuffer; Inc(SrcBuf, FdspBufferPos);
    BytesToCopy:= FdspBufferSize - FdspBufferPos;
    if BytesToCopy > DataSize then BytesToCopy:= DataSize;
    Move(SrcBuf^, DestBuf^, BytesToCopy);
    Inc(DestBuf, BytesToCopy);
    Inc(Result, BytesToCopy);
    Inc(FdspBufferPos, BytesToCopy);
    Dec(DataSize, BytesToCopy);
  end;
end;


function TdspAudioStream.WriteWave(WriteSize: Integer): Integer;

  procedure WriteData(Size: Integer);
  var
    Data1, Data2: Pointer;
    Data1Size, Data2Size: Longint;
  begin
    if FDSBuffer.Lock(FBufferPos, Size, Data1, Data1Size, Data2, Data2Size) then
    begin
      try
        Read(Data1^, Data1Size);
        FWritePosition := FWritePosition + Data1Size;

        if Data2<>nil then
        begin
          Read(Data2^, Data2Size);
          FWritePosition := FWritePosition + Data2Size;
        end;

        FBufferPos := (FBufferPos + DWORD(Data1Size) + DWORD(Data2Size)) mod FDSBufferSize;
      finally
        FDSBuffer.UnLock;
      end;
    end;
  end;

  procedure WriteData2(Size: Integer);
  var
    Data1, Data2: Pointer;
    Data1Size, Data2Size, s1, s2: Longint;
  begin
    if FDSBuffer.Lock(FBufferPos, Size, Data1, Data1Size, Data2, Data2Size) then
    begin
      try
        s1 := Read(Data1^, Data1Size);
        FWritePosition := FWritePosition + s1;
        FBufferPos := (FBufferPos + DWORD(s1)) mod FDSBufferSize;
        Inc(Result, s1);

        if (Data2<>nil) and (s1=Data1Size) then
        begin
          s2 := Read(Data2^, Data2Size);
          FWritePosition := FWritePosition + s2;
          FBufferPos := (FBufferPos + DWORD(s2)) mod FDSBufferSize;
          Inc(Result, s2);
        end;
      finally
        FDSBuffer.UnLock;
      end;
    end;
  end;

  procedure WriteSilence(Size: Integer);
  var
    C: Byte;
    Data1, Data2: Pointer;
    Data1Size, Data2Size: Longint;
  begin
    if Format^.wBitsPerSample=8 then C := $80 else C := 0;

    if FDSBuffer.Lock(FBufferPos, Size, Data1, Data1Size, Data2, Data2Size) then
    begin
      try
        FillChar(Data1^, Data1Size, C);

        if Data2<>nil then
          FillChar(Data2^, Data2Size, C);
      finally
        FDSBuffer.UnLock;
      end;
      FBufferPos := (FBufferPos + DWORD(Data1Size) + DWORD(Data2Size)) mod FDSBufferSize;
      FWritePosition := FWritePosition + Data1Size + Data2Size;
    end;
  end;

var
  DataSize: Integer;
begin
  if Size>=0 then
  begin
    Result := WriteSize;
    DataSize := Size-FWritePosition;

    if DataSize<=0 then
    begin
      WriteSilence(WriteSize);
    end else
    if DataSize>=WriteSize then
    begin
      WriteData(WriteSize);
    end else
    begin
      WriteData(DataSize);
      WriteSilence(WriteSize-DataSize);
    end;
  end else
  begin
    Result:= 0;
    WriteData2(WriteSize);
  end;
end;


{  TdspAudioFileStream  }

destructor TdspAudioFileStream.Destroy;
begin
  inherited Destroy;
  FWaveFileStream.Free;
end;


procedure TdspAudioFileStream.SetFileName(const Value: string);
begin
  if FFileName=Value then Exit;

  FFileName:= Value;

  if FWaveFileStream<>nil then
  begin
    WaveStream:= nil;
    FWaveFileStream.Free;
    FWaveFileStream:= nil;
  end;

  if Value <> '' then
  begin
    try
      FWaveFileStream:= TWaveFileStream.Create(Value, fmOpenRead or fmShareDenyWrite);
      FWaveFileStream.Open(False);

      if (FWaveFileStream.Format.wBitsPerSample <> 16) then
        raise Exception.Create('This demo supports 16 bits per sample audio only');

      WaveStream:= FWaveFileStream;
    except
      WaveStream:= nil;
      FFileName:= '';
      raise;
    end;
  end;
end;

end.
